library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_misc.all;
use ieee.numeric_std.all;
use work.wic_pkg.all;

entity wic_simulator is
  generic (
    g_SIMULATION : boolean := false;
    g_ENABLE     : boolean := false);
  port (
    rst_n_i         : in  std_logic;
    clk_i           : in  std_logic;
    slot_i          : in  t_wic_slot_out;
    slot_o          : out t_wic_slot_in
  );
end wic_simulator;

architecture rtl of wic_simulator is
  constant c_freq     : integer := 80000000;
  constant c_sim_freq : integer := 200000;
  constant c_ERR_DUR  : integer := 100000;

  signal cnt         : unsigned(26 downto 0);
  signal cnt_dur     : unsigned(16 downto 0);
  signal tick        : std_logic;
  signal err_en      : std_logic;
  signal loops_reg_n : std_logic_vector(c_IOS_SLOT-1 downto 0);
  signal err_inject  : std_logic_vector(c_IOS_SLOT-1 downto 0);
  signal err_loops   : std_logic_vector(c_IOS_SLOT-1 downto 0);
begin

  -- tick once per-second to trigger simulated change in loops
  process(clk_i)
  begin
    if rising_edge(clk_i) then
      if rst_n_i = '0' then
        tick <= '0';
        cnt  <= (others=>'0');
      elsif (g_SIMULATION = true  and cnt < c_sim_freq-1) or
            (g_SIMULATION = false and cnt < c_freq) then
        tick <= '0';
        cnt  <= cnt + 1;
      else
        tick <= '1';
        cnt  <= (others=>'0');
      end if;
    end if;
  end process;

  process(clk_i)
  begin
    if rising_edge(clk_i) then
      if rst_n_i = '0' then
        cnt_dur <= (others=>'0');
        err_en  <= '0';
      else
        if tick = '1' then
          err_en <= '1';
        elsif err_en = '1' and cnt_dur < c_ERR_DUR then
          cnt_dur <= cnt_dur + 1;
        else
          err_en <= '0';
          cnt_dur <= (others=>'0');
        end if;
      end if;
    end if;
  end process;

  -- simulate WIC current loops
  -- bit _n_ is 0 means loop _n_ is closed (no alarm)
  -- take into account requested relays setting and simulated error
  -- injection
  GEN_ERR: if g_ENABLE generate
    slot_o.loops <= slot_i.relays or err_loops;
  end generate;

  GEN_NO_ERR: if not g_ENABLE generate
    slot_o.loops <= slot_i.relays;
  end generate;

  err_loops <= err_inject when(err_en = '1') else
               (others=>'0');

  process(clk_i)
  begin
    if rising_edge(clk_i) then
      if rst_n_i = '0' then
        err_inject(c_IOS_SLOT-1 downto 1) <= (others=>'0');
        err_inject(0) <= '1';
      elsif tick = '1' then
        -- shift left err_inject for the next tick
        --err_inject  <= err_inject(c_IOS_SLOT-2 downto 0) & err_inject(c_IOS_SLOT-1);
        -- shift right err_inject for the next tick
        err_inject  <= err_inject(0) & err_inject(c_IOS_SLOT-1 downto 1);
      end if;
    end if;
  end process;

end rtl;
