//-----------------------------------------------------------------------------
// Title      : Pentair FCM2 simulation model
// Project    : Distributed I/O Tier
//-----------------------------------------------------------------------------
// File       : pentair_fcm2_model.sv
// Author     : Greg Daniluk <grzegorz.daniluk@cern.ch>
// Company    : CERN (BE-CO-HT)
// Standard   : VHDL
//-----------------------------------------------------------------------------
// Description:
// This module is a simulation model for Pentair FCM2 diagnostics module. It
// uses I2C slave implementation from general-cores repository and FCM_interface
// to generate randomized monitoring data.
//-----------------------------------------------------------------------------
//
// Copyright (c) 2018 CERN
//
// Copyright and related rights are licensed under the Solderpad Hardware
// License, Version 0.51 (the “License”) (which enables you, at your option,
// to treat this file as licensed under the Apache License 2.0); you may not
// use this file except in compliance with the License. You may obtain a copy
// of the License at http://solderpad.org/licenses/SHL-0.51.
// Unless required by applicable law or agreed to in writing, software,
// hardware and materials distributed under this License is distributed on an
// “AS IS” BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
// or implied. See the License for the specific language governing permissions
// and limitations under the License.
//
//-----------------------------------------------------------------------------

`include "../fcm2_interface.svh"

module pentair_fcm2_model (clk_i, rst_n_i, sda_b, scl_b);
  input clk_i;
  input rst_n_i;
  inout sda_b;
  inout scl_b;

  reg [7:0] fan_status;
  reg [7:0] temp_status;
  reg [7:0] volt_status;
  reg [0:11] [7:0] fan_speed;

  wire scl_out, scl_en;
  wire sda_out, sda_en;

  FCM_interface ctrl(clk_i, rst_n_i);

  always @(posedge clk_i) begin
    fan_status  = 'hCA;
    temp_status = 'hFE;
    volt_status = 'h01;
    fan_speed   = 'hDEADBEEFBABE010203040506;
  end

  gc_i2c_slave #(
    .g_auto_addr_ack(1))
  i2c_slave (
    .clk_i(clk_i),
    .rst_n_i(rst_n_i),
    .scl_i(scl_b),
    .scl_o(scl_out),
    .scl_en_o(scl_en),
    .sda_i(sda_b),
    .sda_o(sda_out),
    .sda_en_o(sda_en),
    .i2c_addr_i   (ctrl.dev_adr),
    .ack_i        (ctrl.ack),
    .tx_byte_i    (ctrl.tx_byte),
    .rx_byte_o    (ctrl.rx_byte),
    .i2c_sta_p_o  (ctrl.sta_p),
    .i2c_sto_p_o  (ctrl.sto_p),
    .addr_good_p_o(ctrl.adr_good_p),
    .r_done_p_o   (ctrl.r_done),
    .w_done_p_o   (ctrl.w_done)
  );

  assign scl_b = (scl_en)? scl_out : 'bz;
  assign sda_b = (sda_en)? sda_out : 'bz;

  always begin
    ctrl.new_values();
    ctrl.reply();
  end

endmodule
