/*
 * Copyright (C) 2018 CERN
 * Author: Maciej Suminski <maciej.suminski@cern.ch>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef DIOT_CRATE_H
#define DIOT_CRATE_H

#include <nanofip.h>

/**
 * DIOT_Crate represents a CompactPCI-Serial crate stuffed with digital
 * input/output cards (EDA-03828) managed via nanoFIP.
 */
class DIOT_Crate : public nanoFIP
{
public:
    /**
     * Default constructor.
     * @param addr is the FIP agent address.
     * @param cards is the number of I/O cards in the crate.
     */
    DIOT_Crate(int addr, int cards)
        : nanoFIP(addr, VAR_SIZE), card_count(cards) {}

    /**
     * Change an output state.
     * @param slot is the card number (indexing from 0).
     * @param idx is the output number (0 to IO_PER_SLOT - 1 range).
     * @param state is the new output state.
     */
    void SetOutput(int slot, int idx, bool state);

    /**
     * Return an input state.
     * @param slot is the card number (indexing from 0).
     * @param idx is the input number (0 to IO_PER_SLOT - 1 range).
     * @return True when the input is high, false otherwise.
     */
    bool GetInput(int slot, int idx)
    {
        return getConsBit(getDataOffset(slot, idx), getIOMask(idx));
    }

    /**
     * Return an input error state.
     * @param slot is the card number (indexing from 0).
     * @param idx is the input number (0 to IO_PER_SLOT - 1 range).
     * @return True when there is an error, false otherwise.
     */
    bool GetInputErr(int slot, int idx)
    {
        return getConsBit(getInpErrOffset(slot, idx), getIOMask(idx));
    }

    /**
     * Change state of all outputs of a card in a particular slot.
     * @param slot is the card number (indexing from 0).
     * @param state is the new outputs state. The LSB corresponds to I/O number 0.
     */
    void SetSlotOutputs(int slot, uint32_t state);

    /**
     * Return state of all inputs of a card in a particular slot.
     * @param slot is the card number (indexing from 0).
     * @return the inputs state. The LSB corresponds to I/O number 0.
     */
    uint32_t GetSlotInputs(int slot)
    {
        return getConsWord(getDataOffset(slot, 0));
    }

    /**
     * Return state of all input errors of a card in a particular slot.
     * @param slot is the card number (indexing from 0).
     * @return the input errors state. The LSB corresponds to I/O number 0.
     */
    uint32_t GetSlotInputErrs(int slot)
    {
        return getConsWord(getInpErrOffset(slot, 0));
    }

    /**
     * Return the number of I/O cards installed in the crate.
     * Note there is no query mechanism involved, it is just the number
     * that has been passed to the constructor.
     */
    int GetCardCount() const
    {
        return card_count;
    }

    /**
     * Return the number of fans.
     */
    unsigned int GetFanCount() const
    {
        return FAN_COUNT;
    }

    /**
     * Return status of a particular fan.
     * @param fan is the fan number (0 < fan < FAN_COUNT).
     * @return True when functional.
     */
    bool GetFanStatus(unsigned int fan);

    /**
     * Return status of all fans.
     * @return 0 when all fans are functional, else check which bits are
     * asserted to determine which fans are faulty.
     */
    int GetFanStatusAll()
    {
        return GetConsVarData(FANS_STATUS);
    }

    /**
     * Return speed of a particular fan.
     * @param fan is the fan number (0 < fan < FAN_COUNT).
     * @return Fan speed (RPM).
     */
    int GetFanRPM(unsigned int fan);


    /**
     * Return the number of temperature sensors.
     */
    unsigned int GetTempCount() const
    {
        return TEMP_COUNT;
    }


    /**
     * Return status of a particular temperature sensor.
     * @param sensor is the sensor number (0 < sensor < TEMP_COUNT).
     * @return True when functional.
     */
    bool GetTempStatus(unsigned int sensor);

    /**
     * Return status of all temperature sensors.
     * @return 0 when all temperature sensors are functional, else check which
     * bits is asserted to determine which sensors are faulty.
     */
    int GetTempStatusAll()
    {
        return GetConsVarData(TEMP_STATUS);
    }

    /**
     * Return temperature read from a particular sensor.
     * @param sensor is the sensor number (0 < sensor < TEMP_COUNT).
     * @return Temperature (Celsius degrees).
     */
    int GetTemp(unsigned int sensor);


    ///> Voltage status flags
    enum VOLTAGE {
        P3V3    = 0x01,
        P5V0    = 0x02,
        P12V0   = 0x04,
        M12V0   = 0x08,
        INVALID = 0xff
    };

    /**
     * Return power supply status.
     * @return 0 when all voltages are correct, else check which bits are
     * asserted to determine faulty voltages.
     * @see VOLTAGE
     */
    int GetVoltStatusAll()
    {
        return GetConsVarData(VOLT_STATUS);
    }

    /**
     * Return status of a particular votlage.
     * @param volt is the checked voltage.
     * @return True when voltage is correct, false otherwise.
     */
    bool GetVoltStatus(VOLTAGE volt);

    /**
     * Return vector of available voltage enums and their descriptions.
     */
    static const std::vector<std::pair<VOLTAGE, std::string>>& GetVoltages()
    {
        return VOLTAGES;
    }

    /**
     * Return a string describing a voltage.
     */
    static std::string GetVoltName(VOLTAGE volt);

    /**
     * Return voltage enum corresponding to a particular index or INVALID
     * if idx is out of bounds.
     */
    static VOLTAGE GetVoltEnum(unsigned int idx);

private:
    ///> Return pin state offset corresponding to a particular I/O and slot.
    int getDataOffset(int slot, int idx)
    {
        if (slot > card_count) {
            throw std::runtime_error("DIOT_Crate::GetIO(): Too high slot number");
        }

        return slot * (IO_PER_SLOT / 8) + idx / 8;
    }

    ///> Return error status corresponding to a particular input and slot.
    int getInpErrOffset(int slot, int idx)
    {
        if (slot > card_count) {
            throw std::runtime_error("DIOT_Crate::GetIO(): Too high slot number");
        }

        return SLOT1_IN_ERR + slot * (IO_PER_SLOT / 8) + idx / 8;
    }

    ///> Calculates bitmask for particular I/O.
    static uint8_t getIOMask(int idx)
    {
        return (1 << (idx % 8));
    }

    ///> Returns state of a bit of a consumed variable.
    bool getConsBit(int address, uint8_t mask)
    {
        return GetConsVarData(address) & mask;
    }

    ///> Returns a 32-bit word made of 4 consecutive consumed variables.
    uint32_t getConsWord(int address);

    ///> Produced/consumed variable size (expressed in bytes)
    static constexpr int VAR_SIZE = 124;

    ///> Number of I/Os per slot (in the memory map)
    static constexpr int IO_PER_SLOT = 32;

    ///> Number of I/O cards installed in the crate
    const int card_count;

    ///> Number of fans
    static constexpr unsigned int FAN_COUNT = 3;

    ///> Number of temperature sensors
    static constexpr unsigned int TEMP_COUNT = 2;

    ///> Vector storing all voltages
    static const std::vector<std::pair<VOLTAGE, std::string>> VOLTAGES;

    ///> Produced variable map
    enum PRODUCED_VAR_MAP {
        SLOT1_OUT_STATE = 0,
        SLOT2_OUT_STATE = 4,
        SLOT3_OUT_STATE = 8,
        SLOT4_OUT_STATE = 12,
        SLOT5_OUT_STATE = 16,
        SLOT6_OUT_STATE = 20,
        SLOT7_OUT_STATE = 24,
        SLOT8_OUT_STATE = 28
    };

    ///> Consumed variable map
    enum CONSUMED_VAR_MAP {
        SLOT1_IN_STATE = 0,
        SLOT2_IN_STATE = 4,
        SLOT3_IN_STATE = 8,
        SLOT4_IN_STATE = 12,
        SLOT5_IN_STATE = 16,
        SLOT6_IN_STATE = 20,
        SLOT7_IN_STATE = 24,
        SLOT8_IN_STATE = 28,

        SLOT1_IN_ERR = 32,
        SLOT2_IN_ERR = 36,
        SLOT3_IN_ERR = 40,
        SLOT4_IN_ERR = 44,
        SLOT5_IN_ERR = 48,
        SLOT6_IN_ERR = 52,
        SLOT7_IN_ERR = 56,
        SLOT8_IN_ERR = 60,

        FANS_STATUS  = 64,
        TEMP_STATUS  = 65,
        VOLT_STATUS  = 66,

        FAN0_SPEED   = 67,
        FAN1_SPEED   = 69,
        FAN2_SPEED   = 71,

        TEMPERATURE0 = 73,
        TEMPERATURE1 = 74,
        TEMPERATURE2 = 75,
        TEMPERATURE3 = 76,
        TEMPERATURE4 = 77,
        TEMPERATURE5 = 78
    };
};

#endif /* DIOT_CRATE_H */
