/*
 * Copyright (C) 2018 CERN
 * Author: Maciej Suminski <maciej.suminski@cern.ch>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * FMC-nanoFIP test program, based on fiptestbed
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
#include <sys/time.h>

#include <masterfip/libmasterfip.h>

static const uint16_t ftb_var_id[2] = {
    0x05, /* produced var */
    0x06, /* consumed var */
};

#define AGENT_COUNT     1
#define VAR_LENGTH      124

/* counter for various errors */
struct ftb_var_error {
    int write_failed; /* writing new value failed */
    int missed; /* missed data (not updated) */
    int bad_data; /* mismatch between write/read data */
    int not_freshed; /* other than 0x5 promptness&significance */
    int not_significant; /* other than 0x5 promptness&significance */
    int tmo; /* time out no reply from agent */
    int hw_err; /* hw bit error in the frame */
    int hw_bsz; /* hw wrong nb of bytes in the frame */
    int hw_ctrl; /* hw bad control byte in the frame */
    int hw_pdu; /* hw bad pdu byte in the frame */
};

struct ftb_ba {
    struct mstrfip_macrocycle *mc;
    struct mstrfip_data *cons_varlist[AGENT_COUNT];/* consumed varlist */
    struct mstrfip_data *prod_varlist[AGENT_COUNT];/* produced varlist */
    struct mstrfip_data *ident_varlist[AGENT_COUNT];/* identification varlist */
};

struct ftb_dev {
    struct mstrfip_dev *dev;
    struct ftb_ba ba;
    struct ftb_var_error var_err[AGENT_COUNT];

    enum mstrfip_bitrate fbus_speed; /* FIP bus speed get from the HW */
    int agent_count;
    int cycle_uslength;
    int enable_aper_wind;
    uint8_t ctrl_byte[AGENT_COUNT][VAR_LENGTH]; /* value to send to agent acquired in the next cycle*/
    int var_length[AGENT_COUNT];
    uint32_t tr_nstime[AGENT_COUNT];
    int error_detected;
} *ftb;

static int read_vars()
{
    struct mstrfip_data *pvar;
    int found_error, res;

    found_error = 0;
    for (int agt_idx = 0; agt_idx < ftb->agent_count; ++agt_idx) {
        pvar = ftb->ba.cons_varlist[agt_idx];
        memset(pvar->buffer, 0x00, VAR_LENGTH);
        mstrfip_var_update(ftb->dev, pvar);

        switch (pvar->status) {
        case MSTRFIP_DATA_OK:
            // compare values
            res = memcmp(pvar->buffer, ftb->ctrl_byte[agt_idx], VAR_LENGTH);
            if (res) {
                printf("%x%x != %x%x\n",
                        pvar->buffer[0], pvar->buffer[1],
                        ftb->ctrl_byte[agt_idx][0], ftb->ctrl_byte[agt_idx][1]);
                ++(ftb->var_err[agt_idx].bad_data);
                found_error = 1;
                continue; /* leave, one error detected */
            }
            break;

        case MSTRFIP_DATA_PAYLOAD_ERROR:
            found_error = 1;
            if (pvar->payload_error & MSTRFIP_FRAME_PAYLOAD_NOT_REFRESH)
                ++(ftb->var_err[agt_idx].not_freshed);
            if (pvar->payload_error & MSTRFIP_FRAME_PAYLOAD_NOT_SIGNIFICANT)
                ++(ftb->var_err[agt_idx].not_significant);
            break;

        case MSTRFIP_DATA_NOT_RECEIVED:
            found_error = 1;
            ++(ftb->var_err[agt_idx].missed);
            break;

        case MSTRFIP_DATA_FRAME_ERROR:
            found_error = 1;
            if (pvar->frame_error & MSTRFIP_FRAME_TMO)
                ++(ftb->var_err[agt_idx].tmo);
            if (pvar->frame_error & MSTRFIP_FRAME_ERR)
                ++(ftb->var_err[agt_idx].hw_err);
            if (pvar->frame_error & MSTRFIP_FRAME_BAD_CTRL)
                ++(ftb->var_err[agt_idx].hw_ctrl);
            if (pvar->frame_error & MSTRFIP_FRAME_BAD_PDU)
                ++(ftb->var_err[agt_idx].hw_pdu);
            if (pvar->frame_error & MSTRFIP_FRAME_BAD_BSZ)
                ++(ftb->var_err[agt_idx].hw_bsz);
        }
    }
    return found_error;
}

static int update_vars()
{
    int agt_idx, res;
    struct mstrfip_data *pvar;

    /* increment setting */
    /* send new setting */
    res = 0;
    for (agt_idx = 0; agt_idx < ftb->agent_count; ++agt_idx) {
        pvar = ftb->ba.prod_varlist[agt_idx];

        if (ftb->ctrl_byte[agt_idx][1] == 127) {
            ftb->ctrl_byte[agt_idx][0] = 0;
            ftb->ctrl_byte[agt_idx][1] = 1;
        }
        else {
            ++(ftb->ctrl_byte[agt_idx][0]);
            ++(ftb->ctrl_byte[agt_idx][1]);
        }

        /* Send new ctrl value: two bytes filled with ctrl_byte value */
        pvar->bsz = ftb->var_length[agt_idx];
        memcpy((void *)pvar->buffer, (void *)ftb->ctrl_byte[agt_idx], VAR_LENGTH);
        res |= mstrfip_var_write(ftb->dev, pvar);

        if (res) {
            ++ftb->var_err[agt_idx].write_failed;
            fprintf(stderr, "mstrfip_var_write error:%d(%s)\n", res, strerror(res));
        }
    }
    return res;
}

void mstrfip_var_handler(struct mstrfip_dev *dev, struct mstrfip_data *pvar,
                struct mstrfip_irq *irq)
{
    int err1 = read_vars();
    int err2 = update_vars();
    printf("var_handler, error = %d%d\n", err1, err2);
}

void mstrfip_prod_var_handler(struct mstrfip_dev *dev, struct mstrfip_data *pvar,
                struct mstrfip_irq *irq)
{
    printf("prod_var_handler\n");
}

void mstrfip_ident_var_handler(struct mstrfip_dev *dev,
                    struct mstrfip_irq *irq)
{
    printf("ident_var_handler\n");
    mstrfip_identlist_update(ftb->dev, ftb->ba.ident_varlist, ftb->agent_count);
}

void mstrfip_error_handler(struct mstrfip_dev *dev, enum mstrfip_error_list error)
{
    printf("error_handler\n");
}

static int ftb_create_per_var(struct ftb_ba *ba)
{
    struct mstrfip_data_cfg var_cfg;
    int agent_idx;

    /* build prod varlist */
    for (agent_idx = 0; agent_idx < AGENT_COUNT; agent_idx++) {
        var_cfg.max_bsz = ftb->var_length[agent_idx];
        var_cfg.id = (0xFF00 & (ftb_var_id[0] << 8)) | (agent_idx + 1);
        var_cfg.flags = MSTRFIP_DATA_FLAGS_PROD;
        var_cfg.mstrfip_data_handler = (agent_idx == AGENT_COUNT - 1) ?
            mstrfip_prod_var_handler : NULL;
        /* set callback on the last produced periodic var */
        ba->prod_varlist[agent_idx] = mstrfip_var_create(ba->mc, &var_cfg);
        if (ba->prod_varlist[agent_idx] == NULL)
            return -1;
    }

    /* build consumed varlist */
    for (agent_idx = 0; agent_idx < AGENT_COUNT; agent_idx++) {
        var_cfg.max_bsz = ftb->var_length[agent_idx];
        var_cfg.id = (0xFF00 & (ftb_var_id[1] << 8)) | (agent_idx + 1);
        var_cfg.flags = MSTRFIP_DATA_FLAGS_CONS;
        /* set callback for the last agent, it will update all variables */
        var_cfg.mstrfip_data_handler = (agent_idx == AGENT_COUNT - 1) ?
            mstrfip_var_handler : NULL;
        ba->cons_varlist[agent_idx] = mstrfip_var_create(ba->mc, &var_cfg);
        if (ba->cons_varlist[agent_idx] == NULL)
            return -1;
    }
    return 0;
}

#if 0
static int ftb_create_ident_var(struct ftb_ba *ba)
{
    int agent_idx, agent_addr;

    /* build identification varlist */
    for (agent_idx = 0; agent_idx < AGENT_COUNT; agent_idx++) {
        agent_addr = agent_idx + 1; /* addr is in the range of [1..25] */
        ba->ident_varlist[agent_idx] =
            mstrfip_ident_var_create(ba->mc, agent_addr);
        if (ba->ident_varlist[agent_idx] == NULL)
            return -1;
    }
    return 0;
}
#endif

int ftb_ba_start(struct ftb_dev *ftb)
{
    int agt_idx, i, res;
    int agt_addr[AGENT_COUNT];

    /* initial control value for each agent */
    for (agt_idx = 0; agt_idx < AGENT_COUNT; ++agt_idx) {
        ftb->ctrl_byte[agt_idx][0] = 2*agt_idx;
        ftb->ctrl_byte[agt_idx][1] = 2*agt_idx + 1;
    }

    if (mstrfip_ba_reset(ftb->dev) < 0) {
        fprintf(stderr, "Reset BA failed: %s\n",
            mstrfip_strerror(errno));
        return -1;
    }

    if (mstrfip_ba_load(ftb->dev, ftb->ba.mc) < 0) {
        fprintf(stderr, "Loading macrocycle failed: %s\n",
            mstrfip_strerror(errno));
        return -1;
    }

    /* get turn around time of each agents */
    for (i = 0; i < AGENT_COUNT; ++i)
        agt_addr[i] = i + 1;

    res = mstrfip_hw_response_time_get(ftb->dev, AGENT_COUNT, agt_addr, ftb->tr_nstime);

    if (res) {
        fprintf(stderr, "Read turn around time failed: %s\n",
            mstrfip_strerror(errno));
    }

    if (mstrfip_ba_start(ftb->dev) < 0) {
        fprintf(stderr, "Start BA failed: %s\n",
            mstrfip_strerror(errno));
        return -1;
    }
    return 0;
}

static int ftb_set_ba_instr(struct ftb_ba *ba)
{
    struct mstrfip_per_var_wind_cfg pwind_cfg = {0};
    struct mstrfip_aper_var_wind_cfg apwind_cfg = {0};

    /* append a periodic window for produced variables */
    pwind_cfg.varlist = ba->prod_varlist; /* periodic var list */
    pwind_cfg.var_count = AGENT_COUNT; /* periodic var count */
    if (mstrfip_per_var_wind_append(ba->mc, &pwind_cfg) < 0)
        return -1;


    /* append a periodic window for consumed varaiables */
    pwind_cfg.varlist = ba->cons_varlist; /*periodic var list */
    pwind_cfg.var_count = AGENT_COUNT; /* periodic var count */
    if (mstrfip_per_var_wind_append(ba->mc, &pwind_cfg) < 0)
        return -1;


    if (ftb->enable_aper_wind) {
        /* append an aperiodic window */
        apwind_cfg.end_ustime = ftb->cycle_uslength - 200; /* end at 10% of cycle */
        apwind_cfg.enable_diag = 1; /* enable diagnostic */
        /* identification variable list that can be requested at run time */
        apwind_cfg.ident_varlist = ba->ident_varlist;
        apwind_cfg.ident_var_count = AGENT_COUNT;
        /* user handler calls back when an ident request has been fully served */
        apwind_cfg.mstrfip_ident_var_handler = mstrfip_ident_var_handler;
        if (mstrfip_aper_var_wind_append(ba->mc, &apwind_cfg) < 0)
            return -1;
    }

    /* append a wait window without padding (silent) */
    if (mstrfip_wait_wind_append(ba->mc, 1, ftb->cycle_uslength) < 0)
        return -1;

    return 0;
}

int cycle_config(struct mstrfip_hw_cfg *hw_cfg, struct mstrfip_sw_cfg *sw_cfg)
{
    if (mstrfip_hw_cfg_set(ftb->dev, hw_cfg) < 0) {
        fprintf(stderr, "Cannot configure HW: %s\n", mstrfip_strerror(errno));
        return -1;
    }

    /* set SW config */
    if (mstrfip_sw_cfg_set(ftb->dev, sw_cfg) < 0) {
        fprintf(stderr, "Cannot configure SW: %s\n", mstrfip_strerror(errno));
        return -1;
    }

    /* create macro cycle */
    ftb->ba.mc = mstrfip_macrocycle_create(ftb->dev);
    if (ftb->ba.mc == NULL) {
        fprintf(stderr, "Create macrocyle failed: %s\n", mstrfip_strerror(errno));
        return -1;
    }

    /* create periodic variable */
    if (ftb_create_per_var(&ftb->ba) < 0 ) {
        fprintf(stderr, "Create periodic variable failed: %s\n", mstrfip_strerror(errno));
        return -1;
    }

    if (ftb_set_ba_instr(&ftb->ba) < 0) {
        fprintf(stderr, "Programming BA failed: %s\n", mstrfip_strerror(errno));
        return -1;
    }

    if (mstrfip_macrocycle_isvalid(ftb->dev, ftb->ba.mc) == 0) {
        fprintf(stderr, "Macrocycle invalid: %s\n", mstrfip_strerror(errno));
        return -1;
    }

    return 0;
}


int main()
{
    int res, signo;
    siginfo_t siginfo;
    sigset_t newmask;
    struct mstrfip_hw_cfg hw_cfg = {0};
    struct mstrfip_sw_cfg sw_cfg = {0};
    int dev_lun = 0;

    ftb = malloc(sizeof(struct ftb_dev));
    if (ftb == NULL) {
        perror("Can't instantiate FTB device\n");
        exit(1);
    }

    memset(ftb, 0, sizeof(struct ftb_dev));

    for (int i = 0; i < AGENT_COUNT; ++i)
        ftb->var_length[i] = VAR_LENGTH;

    ftb->enable_aper_wind = 0;
    ftb->cycle_uslength = 60000;
    ftb->agent_count = AGENT_COUNT;

    res = mstrfip_init();
    if (res) {
        fprintf(stderr, "Cannot init fip library: %s\n", mstrfip_strerror(errno));
        exit(1);
    }

    ftb->dev = mstrfip_open_by_lun(dev_lun);
    if (ftb->dev == NULL) {
        fprintf(stderr, "Cannot open mstrfip dev: %s\n", mstrfip_strerror(errno));
        exit(1);
    }

    /* reset RT app */
    if (mstrfip_rtapp_reset(ftb->dev) < 0) {
        fprintf(stderr, "Cannot reset rt app: %s\n", mstrfip_strerror(errno));
        exit(1);
    }

    /* Get FIP bus speed */
    if (mstrfip_hw_speed_get(ftb->dev, &ftb->fbus_speed)) {
        fprintf(stderr, "Can't get FIP speed: %s. Exit\n", mstrfip_strerror(errno));
        exit(1);
    }

    /* Cycle configuration */
    hw_cfg.enable_ext_trig = 0;
    hw_cfg.enable_int_trig = 0;
    hw_cfg.enable_ext_trig_term = 0;
    hw_cfg.turn_around_ustime = 14;         // 14 us corresponds to 1 Mbit/s
    sw_cfg.irq_thread_prio = 50;
    sw_cfg.diag_thread_prio = 20;
    sw_cfg.mstrfip_error_handler = mstrfip_error_handler;

    if (cycle_config(&hw_cfg, &sw_cfg) < 0) {
        fprintf(stderr, "Cannot init ba: %s\n", mstrfip_strerror(errno));
        exit(1);
    }

    /* time to start ba */
    res = ftb_ba_start(ftb);
    if (res) {
        fprintf(stderr, "ba_start: %d %s\n", res, mstrfip_strerror(res));
        exit(1);
    }

    /* run for ever till a signal is raised */
    sigemptyset(&newmask); /* unblock any signal */
    sigaddset(&newmask, SIGINT);
    sigaddset(&newmask, SIGTERM);
    sigaddset(&newmask, SIGABRT);
    sigaddset(&newmask, SIGFPE);
    sigaddset(&newmask, SIGILL);
    sigaddset(&newmask, SIGUSR1);
    pthread_sigmask(SIG_BLOCK, &newmask, NULL);

    while(1) {
        signo = sigwaitinfo(&newmask, &siginfo);

        if(signo > 0)
        {
            mstrfip_exit();

            /* TO be completed if necessary */
            if (ftb->dev != NULL)
                mstrfip_close(ftb->dev);

            exit(1);
        }
    }

    return 0;
}
